# -*- coding: utf-8 -*-
#
# Fixtures for administration.
#
# ------------------------------------------------


# imports
# -------
import io
import pytest
import logging
from flask import Flask, request, jsonify
from werkzeug.exceptions import NotFound
from flask_sqlalchemy import SQLAlchemy
from flask_occam import Occam, ActionHandler, transactional, validate, optional, paginate
from wtforms import validators, BooleanField

from . import SANDBOX


# application
# -----------
class Config(object):
    ENV = 'testing'
    TESTING = True
    SQLALCHEMY_ECHO = False
    PROPAGATE_EXCEPTIONS = False
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SQLALCHEMY_DATABASE_URI = 'sqlite:///{}/app.db'.format(SANDBOX)
    PLUGIN_DEFAULT_VARIABLE = True


logs = io.StringIO()
logger = logging.getLogger()
logger.setLevel(logging.INFO)
capture = logging.StreamHandler(logs)
app = Flask(__name__)
logger.addHandler(capture)
app.logger.addHandler(capture)
app.config.from_object(Config)
db = SQLAlchemy(app)
occam = Occam(app, db)


# validators
# ----------
boolean = BooleanField('Boolean', [
    validators.DataRequired(),
])


# endpoints
# ---------
@app.route('/items')
class Items(object):

    @paginate(limit=2, total=lambda: Item.count())
    def get(self):
        """
        GET /items

        Docstrings used in testing auto documentation
        """
        items = Item.all(
            limit=request.args['limit'],
            offset=request.args['offset']
        )
        return [x.json() for x in items], 200

    @validate(name=str)
    @transactional
    def post(self):
        """
        POST /items

        Docstrings used in testing auto documentation
        """
        item = Item.create(**request.json)
        return item.json(), 201


@app.route('/items/<id(Item):item>')
class ItemUpdates(object):
    def get(self, item):
        return item.json(), 200

    @validate(
        name=optional(str),
    )
    @transactional
    def put(self, item):
        item.update(**request.json)
        return item.json(), 200

    @transactional
    def delete(self, item):
        item.delete()
        return jsonify(msg='Deleted item'), 204


# models
# ------
class Item(db.Model):
    __tablename__ = 'items'

    # basic
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False, unique=True, index=True)
    archived = db.Column(db.Boolean, default=False)

    def json(self):
        return dict(
            id=self.id,
            name=self.name,
            archived=self.archived
        )


# fixtures
# --------
@pytest.fixture(scope='session')
def items(client):

    # roles
    items = [
        Item.create(name='one'),
        Item.create(name='two'),
        Item.create(name='three')
    ]
    db.session.commit()

    yield items

    return



# -*- coding: utf-8 -*-
#
# Database mixins
#
# ------------------------------------------------


# imports
# -------
from flask import current_app


# helpers
# -------
def current_db():
    return current_app.extensions['sqlalchemy'].db


# mixins
# ------
class ModelMixin(object):
    """
    Example database mixin to be used in extension.
    """

    def __repr__(self):
        display = self.name if hasattr(self, 'name') else self.id
        return '<{}({})>'.format(self.__class__.__name__, display)

    def json(self):
        """
        Return dictionary with model properties. This
        method should be overriden by models to account
        for model-specific nuances in what to include
        in return payloads.
        """
        from sqlalchemy import inspect

        result = {}
        mapper = inspect(self.__class__)
        for column in mapper.attrs:
            result[column.key] = getattr(self, column.key)
        return result

    def commit(self):
        """
        Commit change using session and return item.
        """
        db = current_db()
        db.session.commit()
        return self

    @classmethod
    def get(cls, *args, **filters):
        """
        Get single item using filter_by query.
        """
        if len(args) == 1 and len(filters) == 0:
            filters['id'] = args[0]
        return cls.query.filter_by(**filters).first()

    def update(self, *args, **kwargs):
        """
        Update current item with specified data.
        """
        # normalize inputs
        if len(args) == 1 and isinstance(args[0], dict):
            kwargs.update(args[0])

        # use init method to include param parsing
        obj = self.__class__(**kwargs)

        # set params
        for key in kwargs:
            if hasattr(self, key):
                setattr(self, key, getattr(obj, key))
        del obj

        db = current_db()
        db.session.flush()
        return self

    @classmethod
    def create(cls, *args, **kwargs):
        """
        Create new record using specified arguments.
        """
        self = cls(*args, **kwargs)
        db = current_db()
        db.session.add(self)
        db.session.flush()
        return self

    def delete(self):
        """
        Delete current model object.
        """
        db = current_db()
        db.session.delete(self)
        db.session.flush()
        return

    @classmethod
    def all(cls, limit=None, offset=0):
        """
        Return all data with specified limit and offset
        """
        return cls.find(limit=limit, offset=offset)

    @classmethod
    def find(cls, limit=None, offset=0, **filters):
        """
        Search database with specified limit, offset,
        and filter criteria.
        """
        query = cls.query.filter_by(**filters).offset(offset)
        if limit is not None:
            query = query.limit(limit)
        return query.all()

    @classmethod
    def count(cls):
        """
        Return total number of items in database.
        """
        return cls.query.count()

    @classmethod
    def upsert(cls, *args, **kwargs):
        """
        Upsert specified data into database. If the data
        doesn't exist in the database, it will be created,
        otherwise, the record will be updated. This method
        automatically detects unique keys by which to query
        the database for existing records.

        .. note:: The performance of this could be improved
                  by doing bulk operations for querying and
                  the create/update process.
        """
        from sqlalchemy import inspect

        # parse inputs
        data, multiple = [], True
        if len(kwargs):
            data.append(kwargs)
            multiple = False
        elif len(args) == 1 and isinstance(args[0], (list, tuple)):
            data = args[0]
        else:
            data = args

        # gather unique columns for querying existing data
        unique = []
        mapper = inspect(cls)
        for col in mapper.attrs:
            if hasattr(col, 'columns'):
                if col.columns[0].unique or col.columns[0].primary_key:
                    unique.append(col.key)

        # query for data and create or update
        result = []
        for record in data:

            # query using unique parameters
            params = {k: record[k] for k in unique if k in record}
            item = cls.get(**params) if len(params) else None

            # update if item exists
            if item is not None:
                item.update(**record)

            # create if it doesn't
            else:
                item = cls.create(**record)

            result.append(item)

        return result if multiple else result[0]

    @classmethod
    def load(cls, data, action=None):
        """
        Helper for loading data into application via config file. Using
        the following model definition as an example:

        .. code-block:: python

            class Item(db.Model):
                __tablename__ = 'item'

                # basic
                id = db.Column(db.Integer, primary_key=True)
                name = db.Column(db.String(255), nullable=False, unique=True, index=True)
                archived = db.Column(db.Boolean, default=False)

        You can seed data from the following config file:

        .. code-block:: yaml

            - name: item 1
              archived: True

            - name: item 2
              archived: False

        Into the application using:

        .. code-block:: python

            # via model directly
            User.seed('config.yml')

            # via db
            db.seed.users('config.yml')

        Arguments:
            data (str): File handle or path to config file.
            action (callable): Function to call on each loaded item.
                               Takes single created item as input.
        """
        db = current_db()
        loader = getattr(db.load, cls.__table__.name)
        return loader(data=data, action=action)




# imports
# -------
import re
from werkzeug.routing import BaseConverter
from werkzeug.exceptions import NotFound


# helpers
# -------
MODELS = dict()


def gather_models():
    """
    Inspect sqlalchemy models from current context and set global
    dictionary to be used in url conversion.
    """
    global MODELS

    from flask import current_app, has_app_context
    if not has_app_context():
        return
    if 'sqlalchemy' not in current_app.extensions:
        return

    # inspect current models and add to map
    db = current_app.extensions['sqlalchemy'].db
    for cls in db.Model._decl_class_registry.values():
        if isinstance(cls, type) and issubclass(cls, db.Model):

            # class name
            MODELS[cls.__name__] = cls

            # lowercase name
            MODELS[cls.__name__.lower()] = cls

            # snake_case name
            words = re.findall(r'([A-Z][0-9a-z]+)', cls.__name__)
            if len(words) > 1:
                alias = '_'.join(map(lambda x: x.lower(), words))
                MODELS[alias] = cls
    return


# converters
# ----------
class ModelConverter(BaseConverter):
    """
    For url inputs containing a model identifier, look
    up the model and return the object.

    This method simplifies a lot of the boilerplate needed
    to do model look ups in REST apis.

    Examples:

        .. code-block:: python

            @app.route('/users/<id(User):user>')
            def get_user(user):
                return jsonify(user.json())

    In addition, this class can be inherited and used
    for other custom parameter url converters. For instance,
    here is how you might use it to create a name converter:

    .. code-block:: python

        class NameConverter(ModelConverter):
            __param__ = 'name'

        app.url_map.converters['name'] = NameConverter

        # ... handlers ...

        @app.route('/users/<name(User):user>')
        def get_user(user):
            return jsonify(user.json())

    """
    __param__ = 'id'

    def __init__(self, map, model):
        self.map = map
        self.model = model
        return

    @property
    def models(self):
        global MODELS
        if not MODELS:
            gather_models()
        return MODELS

    def to_python(self, value):
        mapper = self.models

        # make sure model exists
        if self.model not in mapper:
            raise AssertionError(
                'Specified model `{}` in url converter '
                'not part of application models.'.format(self.model))

        # set up class for conversion
        cls = mapper[self.model]

        # search for the object
        model = cls.get(**{self.__param__: value})
        if model is None:
            raise NotFound

        return model

    def to_url(self, value):
        return super(ModelConverter, self).to_url(getattr(value, self.__param__))



# -*- coding: utf-8 -*-
#
# Pytest configuration
#
# ------------------------------------------------


# imports
# -------
import os
import pytest
import logging

from .fixtures import db, app


# fixtures
# --------
from .fixtures import items  ## noqa


# config
# ------
SETTINGS = dict(
    teardown=True,
    echo=False,
)
APP = None
CLIENT = None
logging.basicConfig(level=logging.ERROR)


# plugins
# -------
def pytest_addoption(parser):
    parser.addoption("-N", "--no-teardown", default=False, help="Do not tear down sandbox directory after testing session.")
    parser.addoption("-E", "--echo", action='store_true', default=False, help="Be verbose in query logging.")
    return


def pytest_configure(config):
    global TEARDOWN
    SETTINGS['teardown'] = not config.getoption('-N')
    SETTINGS['echo'] = config.getoption('-E')
    return


@pytest.fixture(autouse=True, scope='session')
def application(request):
    from . import SANDBOX
    global SETTINGS, APP, CLIENT

    # create sandbox for testing
    if not os.path.exists(SANDBOX):
        os.makedirs(SANDBOX)

    # create application
    if SETTINGS['echo']:
        app.config['SQLALCHEMY_ECHO'] = True

    # create default user
    with app.app_context():
        db.drop_all()
        db.create_all()
        yield app

    # teardown sandbox
    if SETTINGS['teardown']:
        import shutil
        shutil.rmtree(SANDBOX)
    return


@pytest.fixture(scope='session')
def client(application):
    global CLIENT
    if CLIENT is not None:
        yield CLIENT
    else:
        with app.test_client() as CLIENT:
            yield CLIENT
    return



# -*- coding: utf-8 -*-
#
# Testing for url handling.
#
# ------------------------------------------------


# imports
# -------
import pytest
from flask import jsonify, request

from flask_occam import transactional, validate, ActionHandler, QueryHandler, UpdateHandler
from .fixtures import app, Item


# handlers
# --------
@app.route('/items/<id(Item):item>/<action>')
class ItemActions(ActionHandler):

    @transactional
    def archive(self, item):
        item.archived = True
        return jsonify(msg='Archived item'), 204

    @transactional
    def unarchive(self, item):
        item.archived = False
        return jsonify(msg='Unarchived item'), 204


@app.route('/items/<id(Item):item>/<resource>')
class ItemPropertyUpdates(UpdateHandler):

    @validate(data=str)
    @transactional
    def name(self, item):
        item.name = request.json['data']
        return jsonify(msg='Updated item name'), 204


@app.route('/items/<id(Item):item>/<query>')
class ItemQueries(QueryHandler):

    def status(self, item):
        return jsonify(archived=item.archived), 200


@app.route('/items/check/<check>')
class ItemCheck(object):
    def get(self, check):
        item = Item.get(id=check)
        if item is None:
            raise NotFound
        return jsonify(msg='Found Item'), 200


# tests
# -----
class TestCRUD(object):

    def test_create(self, client, items):
        # create it
        response = client.post('/items', json=dict(
            name='create'
        ))
        assert response.status_code == 201
        assert response.json['name'] == 'create'

        # read it back
        response = client.get('/items/{}'.format(response.json['id']))
        assert response.status_code == 200
        assert response.json['name'] == 'create'
        return

    def test_read(self, client, items):
        response = client.get('/items/{}'.format(items[0].id))
        assert response.status_code == 200
        assert response.json['name'] == items[0].name
        return

    def test_update(self, client):
        item = Item.create(name='update').commit()

        # update it
        response = client.put('/items/{}'.format(item.id), json=dict(
            name='not update'
        ))
        assert response.status_code == 200
        assert response.json['name'] == 'not update'

        # read it back
        response = client.get('/items/{}'.format(item.id))
        assert response.status_code == 200
        assert response.json['name'] == 'not update'
        return

    def test_delete(self, client):
        item = Item.create(name='delete').commit()

        # make sure it's there
        response = client.get('/items/{}'.format(item.id))
        assert response.status_code == 200
        assert response.json['name'] == 'delete'

        # drop it
        response = client.delete('/items/{}'.format(item.id))
        assert response.status_code == 204

        # check if it's still there
        response = client.get('/items/{}'.format(item.id))
        assert response.status_code == 404
        return


class TestHandlers(object):

    def test_action_handler(self, client, items):
        response = client.get('/items/{}'.format(items[0].id))
        assert response.json['archived'] is False

        response = client.post('/items/{}/archive'.format(items[0].id))
        assert response.status_code == 204

        response = client.get('/items/{}'.format(items[0].id))
        assert response.json['archived'] is True
        return

    def test_query_handler(self, client, items):
        response = client.get('/items/{}/status'.format(items[1].id))
        assert response.status_code == 200
        assert response.json['archived'] is False
        return

    def test_update_handler(self, client, items):
        item = Item.create(name='handler-update').commit()

        # update it
        response = client.put('/items/{}/name'.format(item.id), json=dict(
            data='not handler update'
        ))
        assert response.status_code == 204

        # read it back
        response = client.get('/items/{}'.format(item.id))
        assert response.status_code == 200
        assert response.json['name'] == 'not handler update'
        return

    def test_param_handler(self, client, items):
        response = client.get('/items/check/1')
        assert response.status_code == 200
        return


class TestAutoDocumentation(object):

    def test_get_doc(self, client, items):
        response = client.get('/docs/items')
        assert response.status_code == 200
        assert 'testing auto documentation' in response.data.decode('utf-8')

        response = client.get('/docs/items/1')
        assert response.status_code == 204
        return



# -*- coding: utf-8 -*-
#
# Plugin Setup
#
# ------------------------------------------------


# imports
# -------
import re
import os
import six
import yaml
from functools import wraps
from flask import Flask, Blueprint, Response, jsonify
import types

from .converters import ModelConverter
from .mixins import ModelMixin
from .errors import ValidationError


# config
# ------
METHODS = [
    'GET',
    'HEAD',
    'POST',
    'PUT',
    'DELETE',
    'CONNECT',
    'OPTIONS',
    'TRACE',
    'PATCH'
]
DOCS = {}
MODELS = {}


# helpers
# -------
def document(obj):
    """
    Parse function or method and save documentation
    information in global DOCS data structure for
    auto-serving documentation.

    Args:
        obj (callable): Callable object with docstring.
    """
    global DOCS, METHODS
    if obj.__doc__:
        m = re.search(r"({})\s+(/.+?)\s".format('|'.join(METHODS)), obj.__doc__)
        if m:
            method, endpoint = m.group(1), m.group(2)
            if endpoint not in DOCS:
                DOCS[endpoint] = {}
            DOCS[endpoint][method] = obj.__doc__
    return


def autojsonify(func):
    """
    Automatically jsonify return payload if a dictionary
    is returned from the url handler. This removes some
    boilerplate associated with always needing to wrap
    json with `jsonify`.
    """
    @wraps(func)
    def _(*args, **kwargs):
        ret = func(*args, **kwargs)
        if isinstance(ret, tuple):
            if isinstance(ret[0], (list, tuple, dict)):
                ret = list(ret)
                ret[0] = jsonify(ret[0])
                ret = tuple(ret)
        elif not isinstance(ret, Response):
            ret = jsonify(ret)
        return ret
    return _


def route(self, rule, **options):
    """
    Decorator for registering view function for a given URL rule.
    This ovverrides the default Flask route decorator to be able
    to handle class-based definitions, where endpoints are specified as
    class methods. With this, you can define routes with a function
    like you normally do in Flask:

    .. code-block:: python

        @app.route('/')
        def index():
            return 'Hello World'


    Or, you can define them with a class:

    .. code-block:: python

        @app.route('/test')
        class Test:
            def get(self):
                return 'GET /test', 200
            def post(self):
                return 'POST /test', 201


    This class-based definition of handlers helps with code organization
    and documentation efforts.
    """
    def decorator(obj):

        # function-based definition
        endpoint = options.pop("endpoint", obj.__name__)
        app = self.app if hasattr(self, 'app') else self
        if isinstance(obj, types.FunctionType):
            document(obj)
            app.add_url_rule(rule, endpoint, autojsonify(obj), **options)

        # class-based definition
        else:
            instance = obj()

            # document callable objects
            for name in dir(instance):
                func = getattr(instance, name)
                if name[0] != '_' and callable(func):
                    document(func)

            # parse/add url rules
            methods = options.pop("methods", METHODS)
            for meth in methods:
                handler = meth.lower()
                if hasattr(instance, handler):
                    endpoint = handler + '_' + obj.__name__.lower()
                    options['methods'] = [meth]
                    func = getattr(instance, handler)
                    app.add_url_rule(
                        rule, endpoint, autojsonify(func),
                        **options
                    )

        return obj
    return decorator


Blueprint.route = route


def gather_models():
    """
    Inspect sqlalchemy models from current context and set global
    dictionary to be used in url conversion.
    """
    global MODELS

    from flask import current_app
    if 'sqlalchemy' not in current_app.extensions:
        return

    # inspect current models and add to map
    db = current_app.extensions['sqlalchemy'].db
    for cls in db.Model._decl_class_registry.values():
        if isinstance(cls, type) and issubclass(cls, db.Model):
            MODELS[cls.__name__] = cls
            MODELS[cls.__table__.name] = cls
    return


class DataLoader(object):
    """
    Helper for loading data into application via config file. Using
    the following model definition as an example:

    .. code-block:: python

        class Item(db.Model):
            __tablename__ = 'item'

            # basic
            id = db.Column(db.Integer, primary_key=True)
            name = db.Column(db.String(255), nullable=False, unique=True, index=True)
            archived = db.Column(db.Boolean, default=False)

    You can seed data from the following config file:

    .. code-block:: yaml

        - name: item 1
          archived: True

        - name: item 2
          archived: False

    Into the application using:

    .. code-block:: python

        # via model directly
        User.seed('config.yml')

        # via db
        db.seed.users('config.yml')

    Additionally, this class supports defining multiple types of models
    in config files. If you want to load multiple types of models via
    config, you can use the following syntax in your config:

    .. code-block:: yaml

        Item:
            - name: item 1
              archived: True

        User:
            - name: test user
              email: email@email.com

    And load the data in your application via:

    .. code-block:: python

        db.seed('config.yml')

    """

    def __init__(self, model=None):
        self.model = model
        return

    def __call__(self, data, action=None):

        def load(filename):
            if hasattr(filename, 'read'):
                fi = filename
            else:
                if not os.path.exists(filename):
                    raise FileNotFoundError(filename)
                fi = open(filename, 'r')
            data = yaml.load(fi, Loader=yaml.FullLoader)
            fi.close()
            return data

        # gather models
        global MODELS
        if not MODELS:
            gather_models()

        # normalize inputs
        if isinstance(data, six.string_types):
            data = load(data)
        if not isinstance(data, (list, tuple)):
            data = [data]

        # format list for known model
        if self.model is not None:
            if not isinstance(self.model, six.string_types):
                model = self.model.__name__
            single = True
            create = [{self.model: data}]

        # format list for many embedded models
        else:
            if isinstance(data, dict):
                reformat = True
                for key in data:
                    if key not in MODELS:
                        reformat = False
                        break
                if reformat:
                    data = [{key: data[key]} for key in data]
            if not isinstance(data, (list, tuple)):
                raise AssertionError('Expected data to be in [{model: data}] format.')
            single = False
            create = data

        # iterate through create list and create the items
        db = {}
        for item in create:
            if not isinstance(item, dict):
                raise AssertionError('Invalid format for seed file, expected [{model: data}, ...]')

            # upsert the data
            model = list(item.keys())[0]
            if model not in MODELS:
                raise AssertionError('Model {} could not be found, or config in invalid format.'.format(model))
            results = MODELS[model].upsert(item[model])
            db[model] = results

            # run specified actions
            if action is not None:
                if not isinstance(results, (list, tuple)):
                    results = [results]
                for instance in results:
                    action(instance)

        return db if not single else db[list(db.keys())[0]]

    def __getattr__(self, attr):
        # invalid use case
        if self.model is not None:
            raise AssertionError('Please use db.load or db.load.table for loading data.')

        # gather models
        global MODELS
        if not MODELS:
            gather_models()

        if attr not in MODELS:
            raise AssertionError('Can not load data for table {}. Table does not exist.'.format(attr))

        return DataLoader(model=attr)


# plugin
# ------
class Occam(object):
    """
    Flask extension class for module, which sets up all flask-related
    capabilities provided by the module. This object can be initialized
    directly:

    .. code-block:: python

        from flask import Flask
        from flask_occam import Occam

        app = Flask(__name__)
        occam = Occam(app)

    Or lazily via factory pattern:

    .. code-block:: python

        occam = Occam()
        app = Flask(__name__)
        occam.init_app(app)

    For additional functionality, it is recommended that the extension
    be linked to the Flask-SQLAlchemy extension tied to the application:

    .. code-block:: python

        from flask_sqlalchemy import SQLAlchemy

        db = SQLAlchemy()
        occam = Occam(db)
        app = Flask(__name__)
        db.init_app(app)
        occam.init_app(app)
    """

    def __init__(self, app=None, db=None):
        self.Blueprint = Blueprint

        # arg mismatch
        if app is not None and \
           db is None and \
           not isinstance(app, Flask):
            self.init_db(app)
            app = None

        # proper spec
        if db is not None:
            self.init_db(db)

        # app is specified properly
        if app is not None:
            self.init_app(app)
        return

    def init_app(self, app, db=None):
        """
        Initialize application via lazy factory pattern.

        Args:
            app (Flask): Flask application.
            db (SQAlchemy): Flask SQLAlchemy extension.
        """
        if db is not None:
            self.init_db(db)

        self.app = app
        self.app.route = types.MethodType(route, self.app)
        self.app.config.setdefault('OCCAM_AUTODOC_ENABLED', True)
        self.app.config.setdefault('OCCAM_AUTODOC_PREFIX', '/docs')
        self.app.config.setdefault('OCCAM_LOG_USER_FORMAT', 'user')
        self.app.config.setdefault('OCCAM_LOG_DEFAULT_LEVEL', 'info')
        self.app.url_map.converters['id'] = ModelConverter
        self.app.register_error_handler(ValidationError, ValidationError.handler)

        # gather models for url converters
        @self.app.before_first_request
        def occam_db_init():
            from .converters import gather_models
            gather_models()

        # add auto-documentation if specified
        if self.app.config['OCCAM_AUTODOC_PREFIX']:
            @app.route(self.app.config['OCCAM_AUTODOC_PREFIX'] + '/<path:endpoint>')
            def autodoc(endpoint):
                endpoint = '/' + endpoint
                if endpoint in DOCS:
                    doc = []
                    for method in METHODS:
                        if method in DOCS[endpoint]:
                            doc.append(DOCS[endpoint][method])
                    return "<pre>\n" + '\n---\n'.join(doc) + "\n</pre>\n", 200
                return '', 204

        return

    def init_db(self, db):
        """
        Initialize database model extensions.

        Args:
            db (SQAlchemy): Flask SQLAlchemy extension.
        """
        self.db = db
        self.db.Model.__bases__ += (ModelMixin,)
        self.db.load = DataLoader()
        return
